/*
The GPLv3 License (GPLv3)

Copyright (c) 2022 Akos Horvath

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "client.h"
#include "util.h"
#include "wm.h"
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

// TODO
XWindowChanges wm_client_to_xwchanges(Client *c) 
{

    XWindowChanges ch;
    ch.x = c->x;
    ch.y = c->y;
    ch.width = c->w;
    ch.height = c->h;

    return ch;
}

Client* wm_client_find(Wm *wm, Window w)
{
    PtrArray clients = wm_treenode_find_client_nodes_ptr(wm->smon->workspaces[wm->smon->selws].tree);
    Client *ret = NULL;

    for (size_t i = 0; i < clients.size; i++) {
        Client *c = ((TreeNode*)clients.ptrs[i])->client;
        if (c->window == w) {
            ret = c;
            goto ret;
        }
    }

ret:
    wm_ptrarray_free(&clients);
    return ret;
}

Client* wm_client_create(Wm *wm, Window w)
{
    DEBUG_PRINT("%s\n", __func__);
    XTextProperty xtp;
    int strln;

    XGetWMName(wm->display, w, &xtp);

    Client *c = calloc(1, sizeof(Client));

    c->next = NULL;
    c->prev = NULL;
    c->window = w;
    c->m = wm->smon;
    c->ws = &wm->smon->workspaces[wm->smon->selws];
    c->hidden = true;
    c->is_floating = false;
    c->name = NULL;

    if (xtp.value) {
        strln = strlen((char*)xtp.value);
        DEBUG_PRINT("%s allocating %d bytes for c->name\n", __func__, strln)
        c->name = malloc(strln+1);
        strcpy(c->name, (char*)xtp.value);
    }

    XSelectInput(wm->display, c->window, FocusChangeMask | EnterWindowMask |
                 PointerMotionMask);

    wm_client_set_atom(wm, c, "_NET_WM_DESKTOP", (unsigned char*)&c->ws->index,
                       XA_CARDINAL, 1);


    XFree(xtp.value);
    return c;
}

void wm_client_handle_window_types(Wm *wm, Client *c, XMapRequestEvent e)
{
    DEBUG_PRINT("%s\n", __func__);
    RETURN_IF_NULL(c);

    unsigned char *prop_ret;
    unsigned long nitems_ret;
    Atom type;
    bool is_normal = false;

    type = wm_client_get_atom(wm, c, "_NET_WM_WINDOW_TYPE", &prop_ret, &nitems_ret);

    if (type == ULONG_MAX) {
        fprintf(stderr, "wm_client_get_atom failed\n");
        return;
    }

    for (size_t i = 0; i < nitems_ret; i++) {
        char *atom_name = XGetAtomName(wm->display, ((Atom*)prop_ret)[i]);
        DEBUG_PRINT("ConfigureRequest handler: window %d has property %s\n",
                    (int)e.window, atom_name);
        if (strcmp(atom_name,"_NET_WM_WINDOW_TYPE_NORMAL") == 0) {
            is_normal = true;
        } else if (strcmp(atom_name, "_NET_WM_WINDOW_TYPE_DOCK") == 0) {
            // wm->dock = e.window;
            fprintf(stderr, "client should not have _NET_WM_WINDOW_TYPE_DOCK type\n");
            // todo function to exit
            exit(1);
        } else if (strcmp(atom_name, "_NET_WM_WINDOW_TYPE_DIALOG") == 0) {
            c->is_floating = true;
        }
        XFree(atom_name);
    }

    if (!is_normal) {
        DEBUG_PRINT("configure handler: window %d is not normal\n",
                    (int)e.window)
        // wm_client_kill(wm, c);
        return;
    }

    XFree(prop_ret);
}

void wm_client_hide(Wm *wm, Client *c)
{
    RETURN_IF_NULL(c);

    if (c->hidden)
        return;

    c->hidden = true;
    XUnmapWindow(wm->display, c->window);
}

void wm_client_show(Wm *wm, Client *c)
{
    RETURN_IF_NULL(c);

    if (!c->hidden)
        return;

    c->hidden = false;
    XMapWindow(wm->display, c->window);
}

void wm_client_focus(Wm *wm, Client *c)
{
    DEBUG_PRINT("%s\n", __func__);
    RETURN_IF_NULL(c);
    Client *old_focused_client = wm->focused_client;
    if (!old_focused_client)
        old_focused_client = wm_client_get_focused(wm);

    if (old_focused_client)
        old_focused_client->focused = false;

    XSetInputFocus(wm->display, c->window, RevertToPointerRoot, CurrentTime);
    wm_client_set_atom(wm, c, "_NET_ACTIVE_WINDOW", (unsigned char*) &c->window,
                       XA_WINDOW, 1);
    c->focused = true;
    wm->focused_client = c;
    TreeNode *node = wm_treenode_ptr_find_client_node(c->ws->tree, c);
    c->ws->focused_node = node;
    wm_monitor_clients_border(wm, c->m);
}

void wm_client_focus_dir(Wm *wm, Client *c, int dir)
{
    RETURN_IF_NULL(c);

    switch (dir) {
        case UP:
            wm_client_focus(wm, wm_client_get_dir_rel_c(c, dir));
            break;
        case DOWN:
            wm_client_focus(wm, wm_client_get_dir_rel_c(c, dir));
            break;
        case LEFT:
            wm_client_focus(wm, wm_client_get_dir_rel_c(c, dir));
            break;
        case RIGHT:
            wm_client_focus(wm, wm_client_get_dir_rel_c(c, dir));
            break;
        default:
            fprintf(stderr, "wm: %s invalid direction: %d\n", __func__, dir);
            return;
    }
}

void wm_client_swap_dir(Wm* wm, Client *c, int dir)
{
    RETURN_IF_NULL(c);

    TreeNode *client_node = wm_treenode_ptr_find_client_node(c->ws->tree, c);
    assert(client_node);

    Client *client = wm_client_get_dir_rel_c(c, dir);
    if (!client)
        return;

    TreeNode *client_node1 = wm_treenode_ptr_find_client_node(c->ws->tree, client);

    wm_treenode_swap(c->ws->tree, client_node, client_node1);

    wm_layout(wm, wm->smon);
}

void wm_client_free(Wm *wm, Client *c)
{
    DEBUG_PRINT("%s\n", __func__);
    RETURN_IF_NULL(c);

    if (c->name)
        free(c->name);

    free(c);
}

void wm_client_kill(Wm *wm, Client *c)
{
    DEBUG_PRINT("%s\n", __func__);
    RETURN_IF_NULL(c);
    Monitor *m;

    m = c->m;

    XEvent event;
    Atom delete_atom = XInternAtom(wm->display, "WM_DELETE_WINDOW", False);

    event.type = ClientMessage;
    event.xclient.window = c->window;
    event.xclient.display = wm->display;
    event.xclient.message_type = delete_atom;
    event.xclient.format = 32;
    event.xclient.data.l[0] = delete_atom;

    if (XSendEvent(wm->display, c->window, false, NoEventMask, &event) != Success)
        XKillClient(wm->display, c->window);

    TreeNode* parent = wm_treenode_remove_client(wm, c->ws->tree, c);

    wm_client_free(wm, c);
    wm->focused_client = NULL;

    wm_layout(wm, m);

    if (parent && parent->type == NODE_CLIENT)
        wm_client_focus(wm, parent->client);
}

void wm_client_set_atom(Wm *wm, Client *c, const char* name, const unsigned char *data,
                        Atom type, int nelements)
{
    RETURN_IF_NULL(c);
    RETURN_IF_NULL(name);
    RETURN_IF_NULL(data);

    XChangeProperty(wm->display, c->window, XInternAtom(wm->display, name, False),
                    type, 32, PropModeReplace, data, nelements);
}

// probably not working
Atom wm_client_get_atom(Wm *wm, Client *c, const char *name, unsigned char **atom_ret,
                        unsigned long *nitems_ret)
{
    if (!c)
        return ULONG_MAX;
    if (!name)
        return ULONG_MAX;
    if (!atom_ret)
        return ULONG_MAX;
    if (!nitems_ret)
        return ULONG_MAX;

    Atom type_ret;
    int format_ret;
    unsigned long bytes_remain_ret;
    int ret;

    ret = XGetWindowProperty(wm->display, c->window,
            XInternAtom(wm->display, name, False), 0, (~0L), False,
            AnyPropertyType, &type_ret,
            &format_ret, nitems_ret, &bytes_remain_ret, atom_ret);

    if (ret != Success || (nitems_ret == 0 && type_ret == 0)) {
        fprintf(stderr, "wm: XGetWindowProperty failed\n");
        return ULONG_MAX;
    }

    DEBUG_PRINT("XGetWindowProperty return values, window %d:\n", (int)c->window);
    DEBUG_PRINT("actual type return: %d\n", (int)type_ret);
    DEBUG_PRINT("actual format return: %d\n", format_ret)
    DEBUG_PRINT("actual number of items return: %ld\n", *nitems_ret)
    DEBUG_PRINT("bytes remaining: %ld\n", *nitems_ret)
    for (int i = 0; i < *nitems_ret; i++) {
        char *atom_name = XGetAtomName(wm->display, ((Atom*)*atom_ret)[i]);
        printf("property return str: %s\n", atom_name);
        XFree(atom_name);
    }

    return type_ret;
}

Client* wm_client_get_dir_rel_c(Client *c, int dir)
{
    DEBUG_PRINT("%s c: %p\n", __func__, c)

    if (!c)
        return NULL;

    Client *ret = NULL;
    int x, y, dx, dy;

    switch (dir) {
        case UP:
            x = c->x + c->w/2;
            y = c->y;
            dx = 0;
            dy = -1;
            break;
        case DOWN:
            x = c->x + c->w/2;
            y = c->y + c->h;
            dx = 0;
            dy = 1;
            break;
        case LEFT:
            x = c->x;
            y = c->y + c->h/2;
            dx = -1;
            dy = 0;
            break;
        case RIGHT:
            x = c->x + c->w;
            y = c->y + c->h/3;
            dx = 1;
            dy = 0;
            break;
        default:
            fprintf(stderr, "wm: %s invalid direction: %d\n", __func__, dir);
            return NULL;
    }

    PtrArray client_nodes = wm_treenode_find_client_nodes_ptr(c->ws->tree);

    x+=dx*5;
    y+=dy*5;
    do {
        x+=dx*2;
        y+=dy*2;

        for (size_t i = 0; i < client_nodes.size; i++) {
            Client *r = ((TreeNode**)client_nodes.ptrs)[i]->client;
            if ((x > r->x && x < r->x+r->w) && (y > r->y && y < r->y+r->h)) {
                DEBUG_PRINT("%s ", __func__)
                DEBUG_PRINT("found window %d in direction ", (int)c->window)
                DEBUG_PRINT("%d\n", dir)
                ret = r;
                goto ret;
            }
        }
    // TODO: bar
    } while((x > 0 && x < c->m->info.width) && 
            (y > 0 && y < c->m->info.height));

ret:
    wm_ptrarray_free(&client_nodes);
    DEBUG_PRINT("%s returning %p\n", __func__, ret);
    return ret;
}

Client* wm_client_get_focused(Wm *wm)
{
    Workspace *ws = &wm->smon->workspaces[wm->smon->selws];
    TreeNode *node = wm_treenode_ptr_find_focused_client_node(ws->tree);

    if (!node)
        return NULL;

    return node->client;
}

void wm_client_border(Wm *wm, Client *c)
{
    RETURN_IF_NULL(c);

    XVisualInfo vinfo;
    XColor color;
    XColor fcolor;
    Colormap cmap;

    XMatchVisualInfo(wm->display, DefaultScreen(wm->display), 32,
                     TrueColor, &vinfo);

    cmap = XCreateColormap(wm->display, c->window, vinfo.visual, AllocNone);
    XParseColor(wm->display, cmap, wm->config.focused_border_col, &fcolor);
    XParseColor(wm->display, cmap, wm->config.border_col, &color);
    XAllocColor(wm->display, cmap, &fcolor);
    XAllocColor(wm->display, cmap, &color);

    XSetWindowBorderWidth(wm->display, c->window, wm->config.border_width);

    if (wm_client_is_focused(wm, c)) {
        DEBUG_PRINT("drawing focused border for: %d\n", (int)c->window);
        XSetWindowBorder(wm->display, c->window, fcolor.pixel);
    }
    else
        XSetWindowBorder(wm->display, c->window, color.pixel);
}

void wm_monitor_clients_border(Wm *wm, Monitor *m)
{
    DEBUG_PRINT("%s\n", __func__)
    RETURN_IF_NULL(m);
    PtrArray clients = wm_treenode_find_client_nodes_ptr(m->workspaces[m->selws].tree);
    if (clients.size == 0)
        return;

    for (size_t i = 0; i < clients.size; i++) {
        Client *client = ((TreeNode*)clients.ptrs[i])->client;
        DEBUG_PRINT("monitor border c: %p\n", client)
        DEBUG_PRINT("monitor border c window: %ld\n", client->window)
        wm_client_border(wm, client);
    }

    wm_ptrarray_free(&clients);
}

bool wm_client_is_focused(Wm *wm, Client *c)
{
    if (c == NULL)
        return false;

    Window w;
    int r;

    XGetInputFocus(wm->display, &w, &r);

    if (w == c->window) {
        return true;
    }

    return false;
}
